use crate::{flow::comparator::*, func::*};

use super::{binary::*, *};

struct Bounds<A> {
    l: Option<Rc<A>>,
    r: Option<Rc<A>>,
}

impl<A> Clone for Bounds<A> {
    fn clone(&self) -> Self {
        Self {
            l: self.l.clone(),
            r: self.r.clone(),
        }
    }
}

pub enum BoundsError<A> {
    BoundsViolated { l: Rc<A>, r: Rc<A> },
}

pub enum BoundError<A, E> {
    Avl(AvlError<E>),
    Bounds(BoundsError<A>),
}

impl<A, E> From<AvlError<E>> for BoundError<A, E> {
    fn from(value: AvlError<E>) -> Self {
        Self::Avl(value)
    }
}

impl<A, E> From<BoundsError<A>> for BoundError<A, E> {
    fn from(value: BoundsError<A>) -> Self {
        Self::Bounds(value)
    }
}

pub type BoundErrorA<'a, Ctx, A> = BoundError<A, ParseError<'a, Ctx, Fctr<'a, Ctx, A>>>;

pub type BoundResult<'a, Ctx, A, T> = Result<T, BoundErrorA<'a, Ctx, A>>;

pub type BoundNodeResult<'a, Ctx, A> = BoundResult<'a, Ctx, A, BoundNode<'a, Ctx, A>>;

pub type BoundTreeResult<'a, Ctx, A> = BoundResult<'a, Ctx, A, BoundTree<'a, Ctx, A>>;

pub struct BoundTree<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>> {
    tree: AvlTree<'a, Ctx, A>,
    bounds: Bounds<A>,
}

pub struct BoundNode<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>> {
    node: AvlNode<'a, Ctx, A>,
    bounds: Bounds<A>,
}

pub struct BoundReference<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>> {
    reference: AvlReference<'a, Ctx, A>,
    bounds: Bounds<A>,
}

impl<A> Bounds<A> {
    fn unbound() -> Self {
        Bounds { l: None, r: None }
    }

    fn ordered(
        l: &Rc<A>,
        r: &Rc<A>,
        comparator: &impl Comparator<A>,
    ) -> Result<(), BoundsError<A>> {
        if let Comparison::R = comparator.pick_smaller(l, r) {
            Err(BoundsError::BoundsViolated {
                l: l.clone(),
                r: r.clone(),
            })
        } else {
            Ok(())
        }
    }

    fn new(
        l: Option<Rc<A>>,
        r: Option<Rc<A>>,
        comparator: &impl Comparator<A>,
    ) -> Result<Self, BoundsError<A>> {
        if let (Some(kl), Some(kr)) = (&l, &r) {
            Self::ordered(kl, kr, comparator)?
        }
        Ok(Bounds { l, r })
    }

    fn split(
        self,
        key: &Rc<A>,
        comparator: &impl Comparator<A>,
    ) -> Result<(Self, Self), BoundsError<A>> {
        Ok((
            Self::new(self.l, Some(key.clone()), comparator)?,
            Self::new(Some(key.clone()), self.r, comparator)?,
        ))
    }

    fn join(
        l: Self,
        r: Self,
        key: &Rc<A>,
        comparator: &impl Comparator<A>,
    ) -> Result<Self, BoundsError<A>> {
        if let Some(lr) = &l.r {
            Self::ordered(lr, key, comparator)?
        }
        if let Some(rl) = &r.l {
            Self::ordered(key, rl, comparator)?
        }
        Self::new(l.l, r.r, comparator)
    }
}

pub type BoundSplit<'a, Ctx, A> = (BoundTree<'a, Ctx, A>, BoundTree<'a, Ctx, A>, Rc<A>);

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>> BoundNode<'a, Ctx, A> {
    pub fn into_tree(self) -> BoundTreeResult<'a, Ctx, A> {
        let tree = self.node.into_tree()?;
        let bounds = self.bounds;
        Ok(BoundTree { tree, bounds })
    }

    pub fn new(
        l: BoundTree<'a, Ctx, A>,
        r: BoundTree<'a, Ctx, A>,
        key: Rc<A>,
        comparator: &impl Comparator<A>,
    ) -> BoundNodeResult<'a, Ctx, A> {
        let bounds = Bounds::join(l.bounds, r.bounds, &key, comparator)?;
        let node = AvlNode::new(l.tree, r.tree, key)?;
        Ok(Self { node, bounds })
    }

    pub fn split(
        &self,
        comparator: &impl Comparator<A>,
    ) -> Result<BoundSplit<'a, Ctx, A>, BoundsError<A>> {
        let (bl, br) = self.bounds.clone().split(&self.node.key, comparator)?;
        Ok((
            BoundTree {
                tree: self.node.l.clone(),
                bounds: bl,
            },
            BoundTree {
                tree: self.node.r.clone(),
                bounds: br,
            },
            self.node.key.clone(),
        ))
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>> BoundTree<'a, Ctx, A> {
    pub fn empty(factory: Fctr<'a, Ctx, A>) -> Self {
        Self {
            tree: AvlTree::empty(factory),
            bounds: Bounds::unbound(),
        }
    }

    pub fn reference(&self) -> Option<BoundReference<'a, Ctx, A>> {
        let reference = self.tree.reference()?;
        let bounds = self.bounds.clone();
        Some(BoundReference { reference, bounds })
    }
}

type BoundResolutionError<'a, Ctx, A> =
    ResolutionError<<Ctx as Context<'a>>::LookupError, BoundErrorA<'a, Ctx, A>>;

type BoundResolutionResult<'a, Ctx, A> =
    Result<BoundNode<'a, Ctx, A>, BoundResolutionError<'a, Ctx, A>>;

type BoundResolution<'a, Ctx, A> = Wrapped<'a, Ctx, BoundResolutionResult<'a, Ctx, A>>;

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>> BoundReference<'a, Ctx, A> {
    pub fn resolve(&self) -> BoundResolution<'a, Ctx, A> {
        let bounds = self.bounds.clone();
        <Ctx::T as Functor>::fmap(
            move |resolution| {
                let node = resolution
                    .map_err(|err| match err {
                        ResolutionError::Lookup(lookup) => ResolutionError::Lookup(lookup),
                        ResolutionError::Parse(avl) => ResolutionError::Parse(BoundError::Avl(avl)),
                    })?
                    .as_ref()
                    .clone();
                Ok(BoundNode { node, bounds })
            },
            self.reference.resolve(),
        )
    }
}
